import numpy as np
import pvml
import unittest
import test_data


class TestKNN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_knn_1(self):
        for k in range(1, 5):
            with self.subTest(k):
                X, Y = test_data.non_separable_checkerboard_data_set(48, k)
                Yhat, P = pvml.knn_inference(X, X, Y, k=1)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_knn_5(self):
        for k in range(1, 5):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(48, k)
                Yhat, P = pvml.knn_inference(X, X, Y, k=5)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_knn_auto_k(self):
        for k in range(1, 5):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(24, k)
                k, _ = pvml.knn_select_k(X, Y, maxk=7)
                self.assertGreater(k, 0)
                self.assertLess(k, 8)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import unittest
import test_data


class TestClassificationTree(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train_no_pruning(self):
        tree = pvml.ClassificationTree()
        div = ["gini", "entropy", "error"]
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(50, k)
                tree.train(X, Y, diversity=div[k % 3], pruning_cv=0)
                Yhat = tree.inference(X)[0]
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_train_pruning(self):
        tree = pvml.ClassificationTree()
        div = ["gini", "entropy", "error"]
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.separable_hypercubes_data_set(51, k)
                tree.train(X, Y, diversity=div[k % 3], pruning_cv=5)
                Yhat = tree.inference(X)[0]
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_dump(self):
        tree = pvml.ClassificationTree()
        X, Y = test_data.separable_hypercubes_data_set(21, 3)
        tree.train(X, Y)
        s = tree._dumps()
        self.assertTrue(len(s.splitlines()) > 3)

    def test_check1(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        Y = np.arange(5)
        tree = pvml.ClassificationTree()
        tree.train(X, Y, pruning_cv=0)
        with self.assertRaises(ValueError):
            tree.inference(np.arange(5))

    def test_check2(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        X[:, 0] = 0
        Y = np.arange(5)
        tree = pvml.ClassificationTree()
        tree.train(X, Y, pruning_cv=0)
        with self.assertRaises(ValueError):
            X = np.linspace(-1, 1, 5).reshape(5, 1)
            tree.inference(X)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .checks import _check_size, _check_labels
from .mlp import relu
from .multinomial_logistic import softmax, cross_entropy
from .logistic_regression import sigmoid


class RNN:
    """Recurrent Neural Network."""
    def __init__(self, neuron_counts, cell_type="basic"):
        """Create the RNN."""
        self.neuron_counts = neuron_counts
        connections = zip(neuron_counts[:-2], neuron_counts[1:-1])
        cell_factory = _CELL_TYPES[cell_type.lower()]
        self.cells = [cell_factory(n1, n2) for n1, n2 in connections]
        self.W = (np.random.randn(neuron_counts[-2], neuron_counts[-1]) *
                  np.sqrt(2 / neuron_counts[-2]))
        self.b = np.zeros(neuron_counts[-1])
        self.reset_momentum()

    def reset_momentum(self):
        """Set to zero the accumulators for momentum."""
        self.up_W = np.zeros_like(self.W)
        self.up_b = np.zeros_like(self.b)
        self.up_p = [[np.zeros_like(p) for p in c.parameters()] for c in self.cells]

    def forward(self, X, inits=None):
        """Forward step of the RNN.

        Parameters
        ----------
        X : ndarray, shape (m, t, n)
            sequence of input features (m sequences, t time steps, n components).
        inits : list
            list of initial states, one for each RNN layer (default use zeros).

        Returns
        -------
        P : ndarray, shape (m, t, k)
            sequence of output estimates (m sequences, t time steps, k output estimates).
        """
        m = X.shape[0]
        if inits is None:
            inits = [np.zeros((m, h)) for h in self.neuron_counts[1:-1]]
        H = X
        for cell, init in zip(self.cells, inits):
            H = cell.forward(H, init)
        V = H @ self.W + self.b
        P = self.activation(V)
        self.H = H
        self.P = P
        return P

    def backward(self, Y):
        """Backward step of the RNN.

        Parameters
        ----------
        Y : ndarray, shape (m,) or (m, t)
            target output (integer class labels, one per sequence or one per sequence element).

        Returns
        -------
        D : ndarray, shape (m, t, n)
            derivative of the losses at each time steps with respect to the inputs.
        """
        DV = self.activation_backward(self.P, Y)
        D = DV @ self.W.T
        for cell in self.cells[::-1]:
            D = cell.backward(D)
        self.DV = DV
        return D

    def train(self, X, Y, lr=1e-4, lambda_=1e-5, momentum=0.99,
              steps=10000, batch=None):
        """Train the network.

        Apply multiple steps of backpropagation to train the network.

        Parameters
        ----------
        X : ndarray, shape (m, t, n)
            input sequences (m sequences of length t and with n features).
        Y : ndarray, shape (m,) or (m, t)
            target output (integer class labels, one per sequence or one per sequence element).
        lr : float
            learning rate.
        lambda_ : float
            regularization coefficients.
        momentum : float
            momentum coefficient.
        steps : int
            training iterations.
        batch : int or None
            size of the minibatch used in each step.  When None all
            the data is used in each step.
        """
        _check_size(("mtn, m" if Y.ndim == 1 else "mtn, mt"), X, Y)
        Y = _check_labels(Y, self.W.shape[1])
        m = X.shape[0]
        if batch is None:
            batch = X.shape[0]
        i = m
        indices = np.arange(m)
        for step in range(steps):
            if i + batch > m:
                i = 0
                np.random.shuffle(indices)
            self.backpropagation(X[indices[i:i + batch], ...],
                                 Y[indices[i:i + batch], ...],
                                 lr=lr,
                                 lambda_=lambda_,
                                 momentum=momentum)
            i += batch

    def backpropagation(self, X, Y, lr=1e-4, lambda_=1e-5, momentum=0.99):
        """Backpropagation algorithm

        Parameters
        ----------
        X : ndarray, shape (m, t, n)
            input sequences (a batch of m sequences of length t and with n features).
        Y : ndarray, shape (m,) or (m, t)
            target output (integer class labels, one per sequence or one per sequence element).
        lr : float
            learning rate.
        lambda_ : float
            regularization coefficients.
        momentum : float
            momentum coefficient.
        """
        self.forward(X)
        self.backward(Y)
        h, k = self.W.shape
        grad_W = self.H.reshape(-1, h).T @ self.DV.reshape(-1, k) + lambda_ * self.W
        grad_b = self.DV.sum((0, 1))
        grads = [cell.parameters_grad() for cell in self.cells]

        self.up_W *= momentum
        self.up_W -= lr * grad_W
        self.W += self.up_W
        self.up_b *= momentum
        self.up_b -= lr * grad_b
        self.b += self.up_b
        for cell, grad, up in zip(self.cells, grads, self.up_p):
            for u, p, g in zip(up, cell.parameters(), grad):
                u *= momentum
                u -= lr * g
                p += u

    def save(self, filename):
        """Save the network to the file."""
        cell_params = [p for cell in self.cells for p in cell.parameters()]
        np.savez(filename, neuron_counts=self.neuron_counts, W=self.W, b=self.b, *cell_params)

    @classmethod
    def load(cls, filename):
        """Create a new network from the data saved in the file."""
        data = np.load(filename)
        network = cls(data["neuron_counts"])
        network.W[...] = data["W"]
        network.b[...] = data["b"]
        index = 0
        for cell in network.cells:
            for p in cell.parameters():
                p[...] = data["arr_" + str(index)]
                index += 1
        return network

    def activation(self, V):
        """Final activation function."""
        m, t, n = V.shape
        return softmax(V.reshape(-1, n)).reshape(m, t, n)

    def activation_backward(self, P, Y):
        """Derivatives of the final activation function."""
        if Y.ndim == 1:
            D = np.zeros_like(P)
            D[:, -1, :] = P[:, -1, :]
            # Implicitly subtract the one-hot vectors
            D[np.arange(Y.size), -1, Y] -= 1
        else:
            D = P.reshape(Y.size, -1).copy()
            # Implicitly subtract the one-hot vectors
            D[np.arange(Y.size), Y.reshape(-1)] -= 1
            D = D.reshape(*P.shape)
        return D / Y.size

    def loss(self, Y, P):
        """Compute the average cross-entropy."""
        if Y.ndim == 1:
            return cross_entropy(Y, P[:, -1, :])
        else:
            return cross_entropy(Y.reshape(-1), P.reshape(Y.size, -1))


class RNNAbstractCell:
    """A cell for building RNNs."""

    def forward(self, X, Hinit):
        """Forward step: return the hidden state at each time step.

        Parameters
        ----------
        X : ndarray, shape (m, t, n)
            sequence of input features (m sequences, t time steps, n components).
        Hinit : ndarray, shape (m, h)
            intial state (one vector for each input sequence).

        Returns
        -------
        H : ndarray, shape (m, t, h)
            sequence of hidden states (m sequences, t time steps, h units).
        """
        raise NotImplementedError

    def backward(self, DL):
        """Backward step: return derivatives computed for all time steps.

        Parameters
        ----------
        DL : ndarray, shape (m, t, h)
            derivative of the losses at each time steps with respect to the corresponding states.

        Returns
        -------
        DX : ndarray, shape (m, t, n)
            derivative of the losses at each time steps with respect to the inputs.
        """
        raise NotImplementedError

    def parameters(self):
        """List of parameters of the cell.

        Returns
        -------
        params : list
            list of arrays representing the parameters of the cell.
        """
        raise NotImplementedError

    def parameters_grad(self):
        """Derivative of the total loss with respect to the parameters of the cell.

        Returns
        -------
        params : list
            list of arrays representing the derivatives of the parameters of the cell
            in an order consistent with that of the 'parameters' method.
        """
        raise NotImplementedError


class RNNBasicCell(RNNAbstractCell):
    """A Basic RNN cell."""
    def __init__(self, input_size, hidden_size):
        """Initialize the cell.

        Parameters
        ----------
        input_size : int
            number of input neurons.
        hidden_size : int
            number of hidden neurons.
        """
        # Parameters
        self.W = np.random.randn(input_size, hidden_size) * np.sqrt(2 / input_size)
        self.U = np.random.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        self.b = np.zeros(hidden_size)
        # State
        self.X = np.empty((0, 0, input_size))
        self.Z = np.empty((0, 0, hidden_size))
        self.H = np.empty((0, 0, hidden_size))

    def forward(self, X, Hinit):
        # Given X:
        #   Z[t] = W X[t] + U H[t - 1] + b
        #   H[t] = a(Z[t])
        _check_size("mtn, mh, nh", X, Hinit, self.W)
        m, t, n = X.shape
        X1 = X @ self.W + self.b
        H = np.empty_like(X1)
        Z = np.empty_like(X1)
        for i in range(0, t):
            Z[:, i, :] = X1[:, i, :] + Hinit @ self.U
            H[:, i, :] = self.forward_activation(Z[:, i, :])
            Hinit = H[:, i, :]
        self.X = X
        self.H = H
        self.Z = Z
        return H

    def backward(self, DL):
        # Given DL[t] = dLoss[t] / dH[t]:
        #   DH[t] = DL[t] + U^T DZ[t + 1]
        #   DZ[t] = DH[t] * a'(Z[t])
        _check_size("mth, mth", DL, self.H)
        t = DL.shape[1]
        DH = DL.copy()
        DZ = np.empty_like(DL)
        for i in range(t - 1, -1, -1):
            if i < t - 1:
                DH[:, i, :] += DZ[:, i + 1, :] @ self.U.T
            DZ[:, i, :] = DH[:, i, :] * self.backward_activation(self.H[:, i, :])
        DX = DZ @ self.W.T
        self.DH = DH
        self.DZ = DZ
        self.DX = DX
        return DX

    def forward_activation(self, X):
        """Activation function."""
        return relu(X)

    def backward_activation(self, A):
        """Derivative of the activation, given the activation values."""
        return (A > 0).astype(float)

    def parameters(self):
        return (self.W, self.U, self.b)

    def parameters_grad(self):
        # dTotalLoss / dW = sum( DZ[t] X[t]^T )
        # dTotalLoss / dU = sum( DZ[t] H[t - 1]^T )
        # dTotalLoss / db = sum( DZ[t] )
        n, h = self.W.shape
        DW = self.X.reshape(-1, n).T @ self.DZ.reshape(-1, h)
        DU = self.H[:, :-1, :].reshape(-1, h).T @ self.DZ[:, 1:, :].reshape(-1, h)
        Db = self.DZ.sum(1).sum(0)
        return (DW, DU, Db)


class GRUCell(RNNAbstractCell):
    """Gated Recurring Unit cell."""

    def __init__(self, input_size, hidden_size):
        """Initialize the cell.

        Parameters
        ----------
        input_size : int
            number of input neurons.
        hidden_size : int
            number of hidden neurons.
        """
        # Parameters
        self.Wz = np.random.randn(input_size, hidden_size) * np.sqrt(2 / input_size)
        self.Uz = np.random.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        self.bz = np.zeros(hidden_size)
        self.Wr = np.random.randn(input_size, hidden_size) * np.sqrt(2 / input_size)
        self.Ur = np.random.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        self.br = np.zeros(hidden_size)
        self.Wh = np.random.randn(input_size, hidden_size) * np.sqrt(2 / input_size)
        self.Uh = np.random.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        self.bh = np.zeros(hidden_size)
        # Cell state variables
        self.X = np.empty((0, 0, input_size))
        self.R = np.empty((0, 0, hidden_size))
        self.Sr = np.empty((0, 0, hidden_size))
        self.Z = np.empty((0, 0, hidden_size))
        self.Sz = np.empty((0, 0, hidden_size))
        self.Hnew = np.empty((0, 0, hidden_size))
        self.Sh = np.empty((0, 0, hidden_size))
        self.H = np.empty((0, 0, hidden_size))

    def forward(self, X, Hinit):
        #   Sz[t] = Wz X[t] + Uz H[t - 1] + bz
        #   Z[t] = sigmoid(Sz[t])
        #   Sr[t] = Wr X[t] + Ur H[t - 1] + br
        #   R[t] = sigmoid(Sr[t])
        #   Sh[t] = Wh X[t] + Uh (R[t] * H[t - 1]) + bh
        #   H'[t] = a(Sh[t])
        #   H[t] = (1 - Z[t]) * H[t - 1] + Z[t] * H'[t]
        _check_size("mtn, mh, nh", X, Hinit, self.Wz)
        m, t, n = X.shape
        Xz = X @ self.Wz + self.bz
        Xr = X @ self.Wr + self.br
        Xh = X @ self.Wh + self.bh
        H = np.empty_like(Xz)
        R = np.empty_like(H)
        Sr = np.empty_like(H)
        Z = np.empty_like(H)
        Sz = np.empty_like(H)
        Hnew = np.empty_like(H)
        Sh = np.empty_like(H)
        self.Hinit = Hinit.copy()
        for i in range(0, t):
            Sz[:, i, :] = Xz[:, i, :] + Hinit @ self.Uz
            Z[:, i, :] = sigmoid(Sz[:, i, :])
            Sr[:, i, :] = Xr[:, i, :] + Hinit @ self.Ur
            R[:, i, :] = sigmoid(Sr[:, i, :])
            Sh[:, i, :] = Xh[:, i, :] + (R[:, i, :] * Hinit) @ self.Uh
            Hnew[:, i, :] = self.forward_activation(Sh[:, i, :])
            Hinit = (1 - Z[:, i, :]) * Hinit + Z[:, i, :] * Hnew[:, i, :]
            H[:, i, :] = Hinit
        self.X = X.copy()
        self.R = R
        self.Sr = Sr
        self.Z = Z
        self.Sz = Sz
        self.Hnew = Hnew
        self.Sh = Sh
        self.H = H
        return H

    def backward(self, DL):
        # Given DL[t] = dLoss[t] / dH[t]:
        #   DH[t] = DL[t] + Uz^T DSz[t + 1] + Ur^T DSr[t + 1]
        #         + (Uh^T DSz[t + 1]) * R[t + 1] + DH[t + 1] * (1 - Z[t + 1])
        #
        #   DH'[t] = DH[t] * Z[t]
        #   DSh[t] = DH'[t] * a'(H[t])
        #   DZ[t] = DH[t] * (H'[t] - H[t - 1])
        #   DSz[t] = DZ[t] * Z[t] * (1 - Z[t])
        #   DR[t] = (Uh^T DSh[t]) H[t - 1]
        #   DSr[t] = DR[t] * R[t] * (1 - R[t])
        #
        #   DX[t] = Wz^T DSz + Wr^T DSr + Wh^T DSh
        _check_size("mth, mth", self.H, DL)
        m, t, h = self.H.shape
        DH = DL.copy()
        DX = np.empty_like(self.X)
        DHnew = np.empty_like(self.H)
        DSh = np.empty_like(self.H)
        DZ = np.empty_like(self.H)
        DSz = np.empty_like(self.H)
        DR = np.empty_like(self.H)
        DSr = np.empty_like(self.H)
        for i in range(t - 1, -1, -1):
            # Compute DH
            if i < t - 1:
                DH[:, i, :] += DSz[:, i + 1, :] @ self.Uz.T
                DH[:, i, :] += DSr[:, i + 1, :] @ self.Ur.T
                DH[:, i, :] += (DSh[:, i + 1, :]  @ self.Uh.T) * self.R[:, i + 1, :]
                DH[:, i, :] += (1 - self.Z[:, i + 1, :]) * DH[:, i + 1, :]
            # Update the deratives of gates and logits
            DHnew[:, i, :] = DH[:, i, :] * self.Z[:, i, :]
            DSh[:, i, :] = DHnew[:, i, :] * self.backward_activation(self.Hnew[:, i, :])
            Hold = (self.Hinit if i == 0 else self.H[:, i - 1, :])
            DZ[:, i, :] = DH[:, i, :] * (self.Hnew[:, i, :] - Hold)
            DSz[:, i, :] = DZ[:, i, :] * self.Z[:, i, :] * (1 - self.Z[:, i, :])
            DR[:, i, :] = (DSh[:, i, :]  @ self.Uh.T) * Hold
            DSr[:, i, :] = DR[:, i, :] * self.R[:, i, :] * (1 - self.R[:, i, :])
        # Compute DX
        DX = DSz @ self.Wz.T + DSr @ self.Wr.T + DSh @ self.Wh.T
        self.DH = DH
        self.DX = DX
        self.DHnew = DHnew
        self.DSh = DSh
        self.DZ = DZ
        self.DSz = DSz
        self.DR = DR
        self.DSr = DSr
        return DX

    def forward_activation(self, X):
        """Activation function."""
        return np.tanh(X)

    def backward_activation(self, A):
        """Derivative of the activation, given the activation values."""
        return 1 - (A ** 2)

    def parameters(self):
        """List of parameters of the cell."""
        return (self.Wz, self.Uz, self.bz, self.Wr, self.Ur, self.br, self.Wh, self.Uh, self.bh)

    def parameters_grad(self):
        """Derivative of the total loss with respect to the parameters of the cell."""
        # dTotalLoss / dWz = sum( DSz[t] X[t]^T )
        # dTotalLoss / dUz = sum( DSz[t] H[t - 1]^T )
        # dTotalLoss / dbz = sum( DSz[t] )
        # dTotalLoss / dWr = sum( DSr[t] X[t]^T )
        # dTotalLoss / dUr = sum( DSr[t] H[t - 1]^T )
        # dTotalLoss / dbr = sum( DSr[t] )
        # dTotalLoss / dWh = sum( DSh[t] X[t]^T )
        # dTotalLoss / dUh = sum( DSh[t] (R[t] * H[t - 1])^T )
        # dTotalLoss / dbh = sum( DSh[t] )
        n, h = self.Wz.shape
        Hold = np.concatenate((self.Hinit[:, None, :], self.H[:, :-1, :]), 1)
        Hold = Hold.reshape(-1, h)
        DWz = self.X.reshape(-1, n).T @ self.DSz.reshape(-1, h)
        DWr = self.X.reshape(-1, n).T @ self.DSr.reshape(-1, h)
        DWh = self.X.reshape(-1, n).T @ self.DSh.reshape(-1, h)
        DUz = Hold.T @ self.DSz.reshape(-1, h)
        DUr = Hold.T @ self.DSr.reshape(-1, h)
        DUh = (self.R.reshape(-1, h) * Hold).T @ self.DSh.reshape(-1, h)
        Dbz = self.DSz.sum(1).sum(0)
        Dbr = self.DSr.sum(1).sum(0)
        Dbh = self.DSh.sum(1).sum(0)
        return (DWz, DUz, Dbz, DWr, DUr, Dbr, DWh, DUh, Dbh)


# Factory used to select the cell type from the RNN constructor
_CELL_TYPES = {"basic": RNNBasicCell, "gru": GRUCell}



import numpy as np
import pvml
import unittest
import test_data


class TestPerceptron(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train1(self):
        X, Y = test_data.separable_circle_data_set(50, 2)
        w, b = pvml.perceptron_train(X, Y)
        Yhat, Z = pvml.perceptron_inference(X, w, b)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_train2(self):
        X, Y = test_data.separable_hypercubes_data_set(50, 2)
        w, b = pvml.perceptron_train(X, Y)
        Yhat, Z = pvml.perceptron_inference(X, w, b)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_train3(self):
        X, Y = test_data.non_separable_checkerboard_data_set(51, 2)
        w, b = pvml.perceptron_train(X, Y, steps=100)
        Yhat, Z = pvml.perceptron_inference(X, w, b)
        self.assertGreater((Yhat == Y).sum(), 25)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import io
import unittest
import test_data


class CNNTanh(pvml.CNN):
    """CNN with hyperbolic tangent activation function.

    It is used here because tanh is more reliable than ReLU for
    numerical testing.
    """
    def forward_hidden_activation(self, X):
        """Activation function of hidden layers."""
        return np.tanh(X)

    def backward_hidden_activation(self, Y, d):
        """Derivative of the activation function of hidden layers."""
        # y = tanh(x)  ==>  dy/dx = (1 - tanh(x)^2) = (1 - y^2)
        return d * (1 - Y ** 2)


class TestCNN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_activation_gradient(self):
        """Test the gradient of the activation function."""
        cnn = CNNTanh([1, 1])
        X = np.random.randn(10, 1)
        Y = cnn.forward_hidden_activation(X)
        eps = 1e-7
        Y1 = cnn.forward_hidden_activation(X + eps)
        D = cnn.backward_hidden_activation(Y, np.ones_like(Y))
        D1 = (Y1 - Y) / eps
        error = np.abs(D1 - D).max()
        self.assertAlmostEqual(error, 0.0, 5)

    def test_padding(self):
        channels = [1, 1, 1, 1]
        kernel_sz = [3, 3, 3]
        strides = [1, 1, 1]
        pads = [0, 1, 2]
        cnn = pvml.CNN(channels, kernel_sz, strides, pads)
        X = np.random.randn(1, 11, 11, 1)
        A = cnn.forward(X)
        self.assertEqual(A[0].shape[1], 11)
        self.assertEqual(A[1].shape[1], 9)
        self.assertEqual(A[2].shape[1], 9)
        self.assertEqual(A[3].shape[1], 11)
        Y = np.zeros(X.shape[0], dtype=int)
        D = cnn.backward(Y, A)
        self.assertEqual(A[1].shape[1], D[0].shape[1])
        self.assertEqual(A[2].shape[1], D[1].shape[1])
        self.assertEqual(A[3].shape[1], D[2].shape[1])

    def test_backward(self):
        """Test the gradient of the loss wrt the input."""
        m = 3
        h, w = 7, 9
        n = 8
        k = 2
        channels = [n, 4, 4, k]
        kernel_sz = [3, 3, 3]
        strides = [1, 2, 1]
        pads = [0, 1, 2]

        net = CNNTanh(channels, kernel_sz, strides, pads)
        X = np.random.randn(m, h, w, n)
        Y = np.arange(m) % k
        A = net.forward(X)
        D = net.backward(Y, A)
        grad = net.backward_to_input(D[0], X.shape[1], X.shape[2])
        loss = net.loss(Y, A[-1])
        eps = 1e-7
        for index in np.ndindex(*X.shape):
            backup = X[index]
            X[index] += eps
            with self.subTest(index=index):
                A1 = net.forward(X)
                loss1 = net.loss(Y, A1[-1])
                ratio = (loss1 - loss) / eps
                self.assertAlmostEqual(grad[index], ratio, 5)
            X[index] = backup

    def test_parameter_gradients(self):
        """Test the gradient of the loss wrt the parameters."""
        m = 2
        h, w = 14, 18
        n = 3
        k = 5
        channels = [n, 4, 4, k]
        kernel_sz = [5, 3, 3]
        strides = [2, 1, 2]
        pads = [2, 1, 0]

        net = CNNTanh(channels, kernel_sz, strides, pads)
        X = np.random.randn(m, h, w, n)
        Y = np.arange(m) % k
        A = net.forward(X)
        D = net.backward(Y, A)
        loss = net.loss(Y, A[-1])
        grad_Ws, grad_bs = net.parameters_gradient(A, D)
        eps = 1e-7
        for ps, grad_ps, name in ((net.weights, grad_Ws, "W"), (net.biases, grad_bs, "b")):
            for p, grad_p, l in zip(ps, grad_ps, range(len(ps))):
                for index in np.ndindex(*p.shape):
                    backup = p[index]
                    p[index] += eps
                    with self.subTest(parameter=f"{name}[{l}]", index=index):
                        A1 = net.forward(X)
                        loss1 = net.loss(Y, A1[-1])
                        ratio = (loss1 - loss) / eps
                        self.assertAlmostEqual(grad_p[index], ratio, 5)
                    p[index] = backup

    def test_train(self):
        X, Y = test_data.separable_hypercubes_data_set(50, 2)
        net = pvml.CNN([X.shape[1], 2], [1])
        X = X.reshape(X.shape[0], 1, 1, X.shape[1])
        net.train(X, Y, lr=1e-1, steps=1000)
        Yhat, P = net.inference(X)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_batch(self):
        X, Y = test_data.separable_circle_data_set(10, 2)
        net = pvml.CNN([X.shape[1], 2, 2], [1, 1])
        X = X.reshape(X.shape[0], 1, 1, X.shape[1])
        net.train(X, Y, lr=1e-1, steps=1000, batch=10)
        Yhat, P = net.inference(X)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_saveload(self):
        channels = [2, 3, 4, 5]
        net1 = pvml.CNN(channels)
        f = io.BytesIO()
        net1.save(f)
        f.seek(0)
        net2 = pvml.CNN.load(f)
        for w1, w2 in zip(net1.weights, net2.weights):
            self.assertListEqual(list(w1.ravel()), list(w2.ravel()))
        for b1, b2 in zip(net1.biases, net2.biases):
            self.assertListEqual(list(b1), list(b2))


if __name__ == '__main__':
    unittest.main()



import unittest
import pvml
import numpy as np


class TestNormalization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_meanvar1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.meanvar_normalization(X)
        self.assertAlmostEqual(X1.mean(0)[0], 0)
        self.assertAlmostEqual(X1.mean(0)[1], 0)
        self.assertAlmostEqual(X1.var(0)[0], 1)
        self.assertAlmostEqual(X1.var(0)[1], 1)

    def test_meanvar2(self):
        X = np.arange(100).reshape(10, 10)
        Z = X.copy()
        X1, Z1 = pvml.meanvar_normalization(X, Z)
        self.assertAlmostEqual(np.abs(X1.mean(0)).max(), 0)
        self.assertAlmostEqual(np.abs(X1.var(0) - 1).max(), 0)
        self.assertListEqual(X1.tolist(), Z1.tolist())

    def test_minmax1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.minmax_normalization(X)
        self.assertAlmostEqual(X1.min(0)[0], 0)
        self.assertAlmostEqual(X1.max(0)[1], 1)
        self.assertAlmostEqual(X1.min(0)[0], 0)
        self.assertAlmostEqual(X1.max(0)[1], 1)

    def test_minmax2(self):
        X = np.arange(100).reshape(10, 10)
        Z = X.copy()
        X1, Z1 = pvml.minmax_normalization(X, Z)
        self.assertAlmostEqual(X1.min(0)[0], 0)
        self.assertAlmostEqual(X1.max(0)[1], 1)
        self.assertAlmostEqual(X1.min(0)[0], 0)
        self.assertAlmostEqual(X1.max(0)[1], 1)
        self.assertListEqual(X1.tolist(), Z1.tolist())

    def test_maxabs1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.maxabs_normalization(X)
        self.assertAlmostEqual(np.abs(X1).max(0)[0], 1)
        self.assertAlmostEqual(np.abs(X1).max(0)[1], 1)

    def test_maxabs2(self):
        X = np.arange(100).reshape(10, 10)
        Z = X.copy()
        X1, Z1 = pvml.maxabs_normalization(X, Z)
        self.assertAlmostEqual(np.abs(X1).max(0)[0], 1)
        self.assertAlmostEqual(np.abs(X1).max(0)[1], 1)
        self.assertListEqual(X1.tolist(), Z1.tolist())

    def test_l2_1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.l2_normalization(X)
        self.assertAlmostEqual(np.abs((X1 ** 2).sum(1) - 1).max(), 0)

    def test_l2_2(self):
        X = np.arange(100).reshape(10, 10)
        Z = X.copy()
        X1, Z1 = pvml.l2_normalization(X, Z)
        self.assertAlmostEqual(np.abs((X1 ** 2).sum(1) - 1).max(), 0)
        self.assertListEqual(X1.tolist(), Z1.tolist())

    def test_l1_1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.l1_normalization(X)
        self.assertAlmostEqual(np.abs(np.abs(X1).sum(1) - 1).max(), 0)

    def test_l1_2(self):
        X = np.arange(100).reshape(10, 10)
        Z = X.copy()
        X1, Z1 = pvml.l1_normalization(X, Z)
        self.assertAlmostEqual(np.abs(np.abs(X1).sum(1) - 1).max(), 0)
        self.assertListEqual(X1.tolist(), Z1.tolist())

    def test_whitening1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.whitening(X)
        self.assertAlmostEqual(X1.mean(0)[0], 0)
        self.assertAlmostEqual(X1.mean(0)[1], 0)
        c = np.cov(X1.T)
        self.assertAlmostEqual(c[0, 0], 1)
        self.assertAlmostEqual(c[0, 1], 0)
        self.assertAlmostEqual(c[1, 0], 0)
        self.assertAlmostEqual(c[1, 1], 1)

    def test_whitening2(self):
        X = (np.arange(30) % 4).reshape(10, 3)
        T = np.ones((5, 3))
        X1, T1 = pvml.whitening(X, T)
        self.assertAlmostEqual(np.abs(X1.mean(0)).max(), 0)
        c = np.cov(X1.T)
        self.assertAlmostEqual(np.abs(c - np.eye(3)).max(), 0)
        self.assertEqual(T1.std(0)[0], 0)
        self.assertEqual(T1.std(0)[1], 0)
        self.assertEqual(T1.std(0)[2], 0)

    def test_wrong_dimensions1(self):
        X = np.linspace(0, 1, 10)
        with self.assertRaises(ValueError):
            pvml.meanvar_normalization(X)

    def test_wrong_dimensions2(self):
        X1 = np.linspace(0, 1, 10).reshape(5, 2)
        X2 = np.linspace(0, 1, 12).reshape(4, 3)
        with self.assertRaises(ValueError):
            pvml.meanvar_normalization(X1, X2)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import unittest
import test_data


class TestNaiveBayes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_categorical(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.categorical_data_set(50, k)
                probs, priors = pvml.categorical_naive_bayes_train(X, Y)
                Yhat, scores = pvml.categorical_naive_bayes_inference(X, probs, priors)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_categorical2(self):
        X, Y = test_data.categorical_data_set(50, 2)
        priors = np.array([0.0, 1.0])
        probs, priors = pvml.categorical_naive_bayes_train(X, Y, priors)
        self.assertListEqual(priors.tolist(), [0.0, 1.0])
        Yhat, _ = pvml.categorical_naive_bayes_inference(X, probs, priors)
        self.assertListEqual([1] * Y.size, Yhat.tolist())

    def test_multinomial(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.bow_data_set(50, k)
                w, b = pvml.multinomial_naive_bayes_train(X, Y)
                Yhat, scores = pvml.multinomial_naive_bayes_inference(X, w, b)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_multinomial2(self):
        X, Y = test_data.bow_data_set(50, 2)
        priors = np.array([0.0, 1.0])
        w, b = pvml.multinomial_naive_bayes_train(X, Y, priors)
        Yhat, _ = pvml.multinomial_naive_bayes_inference(X, w, b)
        self.assertListEqual([1] * Y.size, Yhat.tolist())

    def test_gaussian(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.separable_hypercubes_data_set(30, k)
                ms, vs, ps = pvml.gaussian_naive_bayes_train(X, Y)
                Yhat, scores = pvml.gaussian_naive_bayes_inference(X, ms, vs, ps)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_gaussian2(self):
        X, Y = test_data.separable_hypercubes_data_set(50, 2)
        priors = np.array([0.0, 1.0])
        ms, vs, priors = pvml.gaussian_naive_bayes_train(X, Y, priors)
        Yhat, _ = pvml.gaussian_naive_bayes_inference(X, ms, vs, priors)
        self.assertListEqual([1] * Y.size, Yhat.tolist())


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .ksvm import ksvm_train, kernel
from .checks import _check_size, _check_labels


def one_vs_one_ksvm_inference(X, Xtrain, alpha, b, kfun, kparam):
    """Multiclass kernel SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtrain : ndarray, shape (t, n)
         features used during training (one row per feature vector).
    alpha : ndarray, shape (t, k * (k - 1) // 2)
         matrix of learned coefficients.
    b : ndarray, shape (k * (k - 1) // 2,)
         vector of biases.
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    _check_size("mn, tn, ts, s", X, Xtrain, alpha, b)
    # 1) recover the number of classes from s = 1 + 2 + ... + k
    m = X.shape[0]
    s = b.size
    k = int(1 + np.sqrt(1 + 8 * s)) // 2
    votes = np.zeros((m, k))
    K = kernel(X, Xtrain, kfun, kparam)
    logits = K @ alpha + b
    bin_labels = (logits > 0)
    # For each pair of classes...
    j = 0
    for pos in range(k):
        for neg in range(pos + 1, k):
            votes[:, pos] += bin_labels[:, j]
            votes[:, neg] += (1 - bin_labels[:, j])
            j += 1
    labels = np.argmax(votes, 1)
    return labels, votes


def one_vs_one_ksvm_train(X, Y, kfun, kparam, lambda_, lr=1e-3, steps=1000,
                          init_alpha=None, init_b=None):
    """Train a multi-class kernel SVM using the one vs. one strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels.
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_alpha : ndarray, shape (m, k * (k - 1) // 2)
        initial coefficient (None for zero initialization)
    init_b : ndarray, shape (k * (k - 1) // 2,)
        initial biases (None for zero initialization)

    Returns
    -------
    alpha : ndarray, shape (m, k * (k - 1) // 2)
        matrix of learned coefficients.
    b : ndarray(k * (k - 1) // 2,)
        learned biases.
    """
    _check_size("mn, m", X, Y)
    Y = _check_labels(Y)
    k = Y.max() + 1
    m, n = X.shape
    alpha = np.zeros((m, k * (k - 1) // 2))
    b = np.zeros(k * (k - 1) // 2)
    j = 0
    # For each pair of classes...
    for pos in range(k):
        for neg in range(pos + 1, k):
            # Build a training subset
            subset = (np.logical_or(Y == pos, Y == neg)).nonzero()[0]
            Xbin = X[subset, :]
            Ybin = (Y[subset] == pos)
            a1 = (None if init_alpha is None else init_alpha[subset, j])
            b1 = (0 if init_b is None else init_b[j])
            # Train the classifier
            abin, bbin = ksvm_train(Xbin, Ybin, kfun, kparam, lambda_, lr=lr,
                                    steps=steps, init_alpha=a1, init_b=b1)
            alpha[subset, j] = abin
            b[j] = bbin
            j += 1
    return alpha, b


def one_vs_rest_ksvm_inference(X, Xtrain, alpha, b, kfun, kparam):
    """Multiclass kernel SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtrain : ndarray, shape (t, n)
         features used during training (one row per feature vector).
    alpha : ndarray, shape (t, k)
         matrix of learned coefficients.
    b : ndarray, shape (k,)
         vector of biases.
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    _check_size("mn, tn, tk, k", X, Xtrain, alpha, b)
    K = kernel(X, Xtrain, kfun, kparam)
    logits = K @ alpha + b
    labels = np.argmax(logits, 1)
    return labels, logits


def one_vs_rest_ksvm_train(X, Y, kfun, kparam, lambda_, lr=1e-3, steps=1000,
                           init_alpha=None, init_b=None):
    """Train a multi-class kernel SVM using the one vs. rest strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels.
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_alpha : ndarray, shape (m, k * (k - 1) // 2)
        initial coefficient (None for zero initialization)
    init_b : ndarray, shape (k * (k - 1) // 2,)
        initial biases (None for zero initialization)

    Returns
    -------
    alpha : ndarray, shape (m, k * (k - 1) // 2)
        matrix of learned coefficients.
    b : ndarray(k * (k - 1) // 2,)
        learned biases.
    """
    _check_size("mn, m", X, Y)
    Y = _check_labels(Y)
    k = Y.max() + 1
    m, n = X.shape
    alpha = np.zeros((m, k))
    b = np.zeros(k)
    for c in range(k):
        Ybin = (Y == c)
        a1 = (None if init_alpha is None else init_alpha[:, c])
        b1 = (0 if init_b is None else init_b[c])
        abin, bbin = ksvm_train(X, Ybin, kfun, kparam, lambda_, lr=lr,
                                steps=steps, init_alpha=a1, init_b=b1)
        alpha[:, c] = abin
        b[c] = bbin
    return alpha, b



import numpy as np
import pvml
import unittest
import test_data


class TestKSVM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train1(self):
        X, Y = test_data.separable_circle_data_set(50, 2)
        alpha, b = pvml.ksvm_train(X, Y, "rbf", 0.1, 0, lr=1e-1, steps=1000)
        Yhat, P = pvml.ksvm_inference(X, X, alpha, b, "rbf", 0.1)
        self.assertListEqual(Y.tolist(), Yhat.tolist())
        self.assertListEqual(Yhat.tolist(), (P > 0).tolist())

    def test_train2(self):
        X, Y = test_data.separable_hypercubes_data_set(12, 2)
        alpha, b = pvml.ksvm_train(X, Y, "polynomial", 2, 0, lr=10, steps=1000)
        Yhat, P = pvml.ksvm_inference(X, X, alpha, b, "polynomial", 2)
        self.assertListEqual(Y.tolist(), Yhat.tolist())
        self.assertListEqual(Yhat.tolist(), (P > 0).tolist())

    def test_rbf_kernel(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        K = pvml.kernel(X, X, "rbf", 1)
        self.assertListEqual(np.diag(K).tolist(), [1] * 5)
        evals = np.linalg.eigvalsh(K)
        self.assertGreaterEqual(evals.min(), 0)

    def test_polynomial_kernel(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        K = pvml.kernel(X, X, "polynomial", 2)
        evals = np.linalg.eigvalsh(K)
        self.assertGreaterEqual(evals.min(), 0)

    def test_unknown_kernel(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        with self.assertRaises(ValueError):
            pvml.kernel(X, X, "unknown", 2)

    def test_wrong_dimensions1(self):
        X1 = np.linspace(-1, 1, 10).reshape(5, 2)
        X2 = np.linspace(-1, 1, 12).reshape(4, 3)
        with self.assertRaises(ValueError):
            pvml.kernel(X1, X2, "polynomial", 2)

    def test_wrong_dimensions2(self):
        X1 = np.linspace(-1, 1, 10)
        X2 = np.linspace(-1, 1, 10)
        with self.assertRaises(ValueError):
            pvml.kernel(X1, X2, "polynomial", 2)

    def test_wrong_coefficients1(self):
        X = np.linspace(-1, 1, 10)
        alpha = np.linspace(-1, 1, 10)
        b = 0
        with self.assertRaises(ValueError):
            pvml.ksvm_inference(X, X, alpha, b, "polynomial", 2)

    def test_wrong_coefficients2(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        alpha = np.linspace(-1, 1, 10)
        b = 0
        with self.assertRaises(ValueError):
            pvml.ksvm_inference(X, X, alpha, b, "polynomial", 2)

    def test_wrong_coefficients3(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        alpha = np.linspace(-1, 1, 10).reshape(5, 2)
        b = 0
        with self.assertRaises(ValueError):
            pvml.ksvm_inference(X, X, alpha, b, "polynomial", 2)

    def test_wrong_bias(self):
        X = np.linspace(-1, 1, 10).reshape(5, 2)
        alpha = np.linspace(-1, 1, 5)
        b = np.linspace(-1, 1, 5)
        with self.assertRaises(ValueError):
            pvml.ksvm_inference(X, X, alpha, b, "polynomial", 2)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import pvml.checks
import unittest


class TestLogisticRegression(unittest.TestCase):
    def test_check_size_ok(self):
        pvml.checks._check_size("mn, nm, mm", np.empty((3, 5)), np.empty((5, 3)), np.empty((3, 3)))

    def test_check_optional(self):
        pvml.checks._check_size("mn?, mn?, mn?", None, np.empty((2, 4)), None)

    def test_check_arg_error(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_size("mn, nm", np.empty(3,))

    def test_check_dim_error(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_size("mn", np.empty(3,))

    def test_check_size_error(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_size("mn, nm", np.empty((2, 4)), np.empty((2, 4)))

    def test_check_scalar_ok(self):
        pvml.checks._check_size("*", 42)

    def test_check_scalar_error(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_size("*", np.array([42]))

    def test_check_labels_ok(self):
        pvml.checks._check_labels((np.arange(10) % 4).astype(float), 4)

    def test_check_labels_error1(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_labels((np.arange(10) % 4) + 0.5)

    def test_check_labels_error2(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_labels(np.arange(5, -5, -1))

    def test_check_labels_error3(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_labels(np.arange(10) % 4, 3)

    def test_check_categorical_ok(self):
        x = (np.arange(10).reshape(5, 2) % 4).astype(float)
        pvml.checks._check_categorical(x)

    def test_check_categorical_error1(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_categorical((np.arange(12).reshape(4, 3) % 4) + 0.5)

    def test_check_categorical_error2(self):
        with self.assertRaises(ValueError):
            pvml.checks._check_categorical(np.arange(5, -5, -1))


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import unittest
import test_data


class TestLogisticRegression(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train_l2(self):
        X, Y = test_data.separable_hypercubes_data_set(50, 2)
        w, b = pvml.logreg_train(X, Y, 0.0001, lr=10, steps=1000)
        P = pvml.logreg_inference(X, w, b)
        Yhat = (P > 0.5).astype(int)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_train_l1(self):
        X, Y = test_data.separable_stripes_data_set(50, 2)
        w, b = pvml.logreg_l1_train(X, Y, 0.0001, lr=10, steps=1000)
        P = pvml.logreg_inference(X, w, b)
        Yhat = (P > 0.5).astype(int)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_cross_entropy1(self):
        Y = np.array([0, 0, 1, 1])
        e = pvml.binary_cross_entropy(Y, Y)
        self.assertEqual(e, 0)

    def test_cross_entropy2(self):
        Y = np.array([0, 0, 1, 1])
        e = pvml.binary_cross_entropy(Y, 1 - Y)
        self.assertTrue(np.isinf(e))


if __name__ == '__main__':
    unittest.main()



import numpy as np


def _check_size(spec, *xs):
    """Check that the arrays respect the specification.

    spec is a string of comma separated specifications.  Each
    specification apply to one element in xs.  The specification i a
    list of letters representing the expected number of dimensions (*
    if a scalar is expected).  If the same letter is used multiple
    times, the the dimensions must match.  A traling '?' makes None a
    valid value.

    """
    ss = list(map(str.strip, spec.split(",")))
    if len(ss) != len(xs):
        msg = "Not enough arguments (expected {}, got {})"
        raise ValueError(msg.format(len(ss), len(xs)))
    dims = {}
    for s, x in zip(ss, xs):
        if s.endswith("?"):
            if x is None:
                continue
            s = s[:-1]
        if s == "*":
            if np.isscalar(x):
                continue
            else:
                raise ValueError("Scalar value expected")
        if len(s) != np.ndim(x):
            msg = "Expected an array of {} dimensions ({} dimensions found)"
            raise ValueError(msg.format(len(s), np.ndim(x)))
        for n, d in enumerate(s):
            k = x.shape[n]
            if d not in dims:
                dims[d] = k
            elif k != dims[d]:
                msg = "Dimensions do not agree (got {} and {})"
                raise ValueError(msg.format(dims[d], k))


def _check_labels(Y, nclasses=None):
    """Check that data can represent class labels."""
    if not np.issubdtype(Y.dtype, np.integer):
        if np.abs(np.modf(Y)[0]).max() > 0:
            raise ValueError("Expected integers")
        Y = Y.astype(np.int32)
    if Y.min() < 0:
        raise ValueError("Labels cannot be negative")
    if nclasses is not None and Y.max() >= nclasses:
        msg = "Invalid labels (maximum is {}, got {})"
        raise ValueError(msg.format(nclasses - 1, Y.max()))
    return Y


def _check_categorical(X):
    """Check that X contain categorical data.

    If needed X is converted to int.
    """
    if not np.issubdtype(X.dtype, np.integer):
        if np.abs(np.modf(X)[0]).max() > 0:
            raise ValueError("Categorical data must be integers")
        X = X.astype(np.int32)
    if X.min() < 0:
        raise ValueError("Categorical data cannot be negative")
    return X



import numpy as np
import pvml
import unittest
import test_data


class TestSVM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train1(self):
        X, Y = test_data.separable_circle_data_set(50, 2)
        w, b = pvml.svm_train(X, Y, 0, lr=1e-1, steps=1000)
        Yhat, P = pvml.svm_inference(X, w, b)
        self.assertListEqual(Y.tolist(), Yhat.tolist())
        self.assertListEqual(Yhat.tolist(), (P > 0).tolist())

    def test_train2(self):
        X, Y = test_data.separable_hypercubes_data_set(50, 2)
        w, b = pvml.svm_train(X, Y, 0.0001, lr=10, steps=1000)
        Yhat, P = pvml.svm_inference(X, w, b)
        self.assertListEqual(Y.tolist(), Yhat.tolist())
        self.assertListEqual(Yhat.tolist(), (P > 0).tolist())

    def test_train3(self):
        X, Y = test_data.separable_stripes_data_set(50, 2)
        w, b = pvml.svm_train(X, Y, 0.0001, lr=10, steps=1000)
        Yhat, P = pvml.svm_inference(X, w, b)
        self.assertListEqual(Y.tolist(), Yhat.tolist())
        self.assertListEqual(Yhat.tolist(), (P > 0).tolist())

    def test_hinge_loss1(self):
        Y = np.array([0, 1])
        Z = np.array([-1, 1])
        loss = pvml.hinge_loss(Y, Z)
        self.assertEqual(loss, 0)

    def test_hinge_loss2(self):
        Y = np.array([0, 1])
        Z = np.array([0, 0])
        loss = pvml.hinge_loss(Y, Z)
        self.assertEqual(loss, 1)

    def test_hinge_loss3(self):
        Y = np.array([0, 1])
        Z = np.array([1, -1])
        loss = pvml.hinge_loss(Y, Z)
        self.assertEqual(loss, 2)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import unittest
import test_data


class TestMultinomialLogistic(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train1(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(50, k)
                w, b = pvml.multinomial_logreg_train(X, Y, 0, lr=1e-1, steps=1000)
                P = pvml.multinomial_logreg_inference(X, w, b)
                Yhat = P.argmax(1)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_train2(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = test_data.separable_hypercubes_data_set(51, k)
                w, b = pvml.multinomial_logreg_train(X, Y, 1e-4, lr=1, steps=1000)
                P = pvml.multinomial_logreg_inference(X, w, b)
                Yhat = P.argmax(1)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_softmax(self):
        Z = np.random.randn(20, 4)
        P = pvml.softmax(Z)
        error = np.abs(P.sum(1) - 1).max()
        self.assertAlmostEqual(0, error)

    def test_one_hot_vectors(self):
        k = 7
        Y = np.arange(20) % k
        H = pvml.one_hot_vectors(Y, k)
        self.assertListEqual([1] * 20, H.sum(1).tolist())
        self.assertListEqual(np.bincount(Y, minlength=k).tolist(), H.sum(0).tolist())

    def test_cross_entropy(self):
        P = np.array([[1, 0, 0]])
        Y = np.array([0])
        ce = pvml.cross_entropy(Y, P)
        self.assertEqual(0, ce)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .checks import _check_size, _check_labels
from .utils import one_hot_vectors, log_nowarn


# Implementation
# ==============
#
# It is similar to CART trees, but without the support for regression,
# categorical variables, and other advanced features.
#
# In training a 'full' tree is grown with a high likelihood of
# overfitting the data.  A diversity criterion is used to split the
# data until it is divided in omogeneous subsets.
#
# Then, cost-complexity pruning is used to prune the tree to the
# 'right' level of complexity.  CC pruning optimize the following
# objective:
#
#   min R(T, lambda) = error(T) + lambda * size(T)
#
# where size(T) is the number of leaves, and error(T) is the training
# error of the tree T.  For lambda = 0 the initial tree is optimal.
# As lambda gorws, eventually it will be better to replace a whole
# subtree with a leaf.  This happen for
#
#   lambda = Delta-error(t) / (size(t) - 1)
#
# where t is a subtree, Delta-error is the increase in training error
# that the pruning of t wuold cause, and size(t) is the number of
# leaves in the subtree.
#
# With the formula above candidate values of lambda are selected.
# K-fold cross-validation is used to pick the one minimizing the
# generalization error.


class ClassificationTree:
    """Classification trees.

    CART-like model of a classification tree for continuous variables.
    """
    def __init__(self):
        """Create a tree with a single node."""
        self._reset(1, 1)

    def inference(self, X):
        """Prediction of the class labels.

        Parameters
        ----------
        X : ndarray, shape (m, n)
            input features (one row per feature vector).

        Returns
        -------
        ndarray, shape (m,)
            predicted labels (one per feature vector).
        ndarray, shape (m, k)
            class probabilities (one per feature vector).
        """
        _check_size("mn", X)
        self._check_features(X)
        J = self._descend(X)
        probs = self.distribution[J, :]
        labels = probs.argmax(1)
        return labels, probs

    def train(self, X, Y, diversity="gini", minsize=1, pruning_cv=5):
        """Train a classification tree.

        Parameters
        ----------
        X : ndarray, shape (m, n)
            training features.
        Y : ndarray, shape (m,)
            training labels.
        diversity : str
            diversity criterion ('gini', 'entropy' or 'error').
        minsize : int
            minimum number of training samples per node during tree growth.
        pruning_cv : int
            number of cross-validation folds used for cost-complexity
            pruning (0 disable pruning).
        """
        _check_size("mn, m", X, Y)
        Y = _check_labels(Y)
        dfun = _DIVERSITY_FUN[diversity]
        m, n = X.shape
        k = Y.max() + 1
        maxleaves = 2 * m // minsize
        self._reset(2 * maxleaves - 1, k)
        H = one_hot_vectors(Y, k)
        self._grow(X, H, 0, dfun, minsize)
        self._trim()
        if pruning_cv != 0:
            self._prune(X, Y, pruning_cv, diversity, minsize)

    def _reset(self, nodes, classes):
        """Initialize the attributes."""
        self.children = np.zeros((nodes, 2), dtype=int)
        self.feature = np.zeros(nodes, dtype=int)
        self.terminal = np.ones(nodes, dtype=np.bool)
        self.threshold = np.zeros(nodes)
        self.distribution = np.ones((nodes, classes))
        self.nodes = 1

    def _descend(self, X):
        """Return the indexes of the tree nodes in which the samples fall."""
        J = np.zeros(X.shape[0], dtype=int)
        while True:
            s = (~self.terminal[J]).nonzero()[0]
            if s.size == 0:
                break
            c = (X[s, self.feature[J[s]]] < self.threshold[J[s]])
            J[s] = self.children[J[s], 1 - c]  # 0 -> left, 1 -> right
        return J

    def _trim(self):
        """Remove unused nodes."""
        n = self.nodes
        self.children = self.children[:n, :]
        self.feature = self.feature[:n]
        self.terminal = self.terminal[:n]
        self.threshold = self.threshold[:n]
        self.distribution = self.distribution[:n, :]

    def _grow(self, X, H, t, dfun, minsize):
        """Grow the tree by trying to split node t."""
        m, k = H.shape
        self.distribution[t, :] = H.mean(0)
        split = _find_split(X, H, dfun, minsize)
        if split is None:
            return
        self.feature[t] = split[0]
        self.threshold[t] = split[1]
        self.terminal[t] = 0
        self.children[t, 0] = self.nodes
        self.children[t, 1] = self.nodes + 1
        self.nodes += 2
        J = (X[:, split[0]] < split[1])
        self._grow(X[J, :], H[J, :], self.children[t, 0], dfun, minsize)
        self._grow(X[~J, :], H[~J, :], self.children[t, 1], dfun, minsize)

    def _prune(self, X, Y, cv, diversity, minsize):
        """Cost-complexity pruning."""
        leaves = self._count_leaves()[~self.terminal]
        errors = self._pruning_errors(X, Y)[~self.terminal] / Y.size
        lambdas = (errors / (leaves - 1))
        lambdas = np.unique(lambdas)
        folds = np.arange(Y.size) % cv
        np.random.shuffle(folds)
        val = self._eval_prune_lambda(X, Y, 0.0, folds, diversity, minsize)
        best = (val, 0.0)
        for lambda_ in lambdas:
            val = self._eval_prune_lambda(X, Y, lambda_, folds, diversity, minsize)
            best = max(best, (val, lambda_))
        self._prune_lambda(X, Y, best[1])

    def _eval_prune_lambda(self, X, Y, lambda_, folds, diversity, minsize):
        """Estimate the accuracy of the tree pruned according to lambda_."""
        Yhat = np.empty_like(Y)
        tree = ClassificationTree()
        for fold in np.unique(folds):
            J = (folds == fold)
            tree.train(X[~J], Y[~J], diversity=diversity, minsize=minsize,
                       pruning_cv=0)
            tree._prune_lambda(X[~J], Y[~J], lambda_)
            Yhat[J] = tree.inference(X[J])[0]
        return (Y == Yhat).mean()

    def _prune_lambda(self, X, Y, lambda_):
        """Prune the tree at the given cost-complexity level."""
        leaves = self._count_leaves()
        errors = self._pruning_errors(X, Y) / Y.size
        lambdas = (errors / np.maximum(1, (leaves - 1)))
        self.terminal |= (lambdas <= lambda_)

    def _count_leaves(self):
        """Return the number of descendent leaves for each node."""
        leaves = np.ones(self.nodes, dtype=int)
        ts = (~self.terminal).nonzero()[0]
        for t in ts[::-1]:
            leaves[t] = leaves[self.children[t, 0]] + leaves[self.children[t, 1]]
        return leaves

    def _pruning_errors(self, X, Y):
        """Return the number of additional errors caused by pruning nodes."""
        klass = self.distribution.argmax(1)
        errors = np.zeros(self.nodes, dtype=int)
        errors[0] = (klass[0] != Y).sum()
        J = np.zeros(X.shape[0], dtype=int)
        while True:
            s = (~self.terminal[J]).nonzero()[0]
            if s.size == 0:
                break
            c = (X[s, self.feature[J[s]]] < self.threshold[J[s]])
            J[s] = self.children[J[s], 1 - c]  # 0 -> left, 1 -> right
            w = (Y[s] != klass[J[s]])
            e = np.bincount(J[s], weights=w, minlength=self.nodes)
            errors += e.astype(int)
        errors2 = errors.copy()
        ts = (~self.terminal).nonzero()[0]
        for t in ts[::-1]:
            errors2[t] = errors2[self.children[t, 0]] + errors2[self.children[t, 1]]
        return errors - errors2

    def _dumps(self, indent=0, node=0):
        """Sting representation of the tree."""
        pre = "{:3d}{} ".format(node, " " * indent)
        if self.terminal[node]:
            klass = self.distribution[node, :].argmax()
            s = pre + "==> {}\n".format(klass)
        else:
            f = self.feature[node]
            t = self.threshold[node]
            s = "".join([pre + "if x[{}] < {}:\n".format(f, t),
                         self._dumps(indent + 4, self.children[node, 0]),
                         "   {} else:\n".format(" " * indent),
                         self._dumps(indent + 4, self.children[node, 1])])
        return s

    def _check_features(self, X):
        if X.shape[1] < self.feature.max() + 1:
            msg = "Expected feature vectors with at least {} components (got {})."
            raise ValueError(msg.format(self.feature.max() + 1, X.shape[1]))


def _find_split(X, H, criterion, minsize):
    """Helper function that finds the best split.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    H : ndarray, shape (m, k)
         one hot vectors for class labels.
    criterion : function
         function mapping probabilities to the real-values objective.
    minsize : int
         minimum size of the subsets in which X is divided by the split.

    Returns
    -------
    int
        feature used to split the data.
    float
        split threshold.
    float
        value of the criterion function for the split.

    The function returns None if no split is able to improve the
    criterion on the whole set.

    """
    m, n = X.shape
    if m < 2 * minsize:
        return None
    p = H.mean(0, keepdims=True)
    start_criterion = criterion(p)
    split = None
    for j in range(n):
        ret = _decision_stump(X[:, j], H, criterion, minsize)
        if ret is not None and (split is None or ret[1] < split[2]):
            split = j, ret[0], ret[1]
    if split is not None and split[2] < start_criterion:
        return split
    else:
        return None


def _entropy(p):
    """Compute entropy, suppressing warnings and preventing errors."""
    logp = log_nowarn(p)
    logp[p == 0] = 0
    return -(p * logp).sum(1)


_DIVERSITY_FUN = {
    "gini": lambda p: (1 - (p ** 2).sum(1)),
    "entropy": _entropy,
    "error": lambda p:  (1 - p.max(1))
}


def _decision_stump(x, h, criterion, minsize):
    """Helper function that finds the best split on a single feature.

    Parameters
    ----------
    x : ndarray, shape (m,)
         input values.
    h : ndarray, shape (m, k)
         one hot vectors for class labels.
    criterion : function
         function mapping probabilities to the real-values objective.
    minsize : int
         minimum size of the subsets in which X is divided by the split.

    Returns
    -------
    float
        split threshold.
    float
        value of the criterion function for the split.

    The function returns None if no split is possible.
    """
    m, k = h.shape
    # 1) sort the data and find candidate splits
    ii = np.argsort(x)
    # possible split points are those where consecutive values are
    # different, except for the first and last minsize positions.
    sp = (x[ii[minsize - 1:m - minsize]] < x[ii[minsize:m - minsize + 1]])
    sp = sp.nonzero()[0] + minsize - 1
    if sp.size == 0:
        return None
    # 2) compute the class distributions at the split points
    counts_lo = h[ii, :].cumsum(0)
    counts_hi = counts_lo[-1:, :] - counts_lo
    p_lo = ((counts_lo[sp, :]) /
            (counts_lo[sp, :].sum(1, keepdims=True)))
    p_hi = ((counts_hi[sp, :]) /
            (counts_hi[sp, :].sum(1, keepdims=True)))
    p_split = sp / m
    # 3) compute the average criterion and select the lowest
    cost = p_split * criterion(p_lo) + (1 - p_split) * criterion(p_hi)
    j = cost.argmin()
    # 4) return the split value (midway the two consecutvie samples)
    # and the corresponding cost
    best = sp[j]
    threshold = (x[ii[best + 1]] + x[ii[best]]) / 2
    return (threshold, cost[j])



import numpy as np


def log_nowarn(x):
    """Compute the logarithm without warnings in case of zeros."""
    with np.errstate(divide='ignore'):
        return np.log(x)


def one_hot_vectors(Y, classes):
    m = Y.shape[0]
    H = np.zeros((m, classes), dtype=int)
    H[np.arange(m), Y] = 1
    return H



import unittest
import pvml
import numpy as np


def normalize_labels(Y):
    # Change labels by permuting them in sych a way that their first
    # occurrences are sorted.  This makes it possible to compare
    # algorithms that are not consistent in assigning labels (like in
    # clustering).
    #
    # (3, 1, 1, 0, 2, 0) -> (0, 1, 1, 2, 3, 2)
    _, index = np.unique(Y, return_index=True)
    old_labels = np.argsort(index)  # e.g. [3, 1, 0, 2]
    new_labels = np.argsort(old_labels)  # e.g. [2, 1, 3, 0]
    return new_labels[Y]


class TestKMeans(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_one_per_class(self):
        for k in range(1, 12):
            with self.subTest(k):
                X = np.random.randn(k, 3)
                centroids = pvml.kmeans_train(X, k)
                Y, _ = pvml.kmeans_inference(X, centroids)
                self.assertListEqual(normalize_labels(Y).tolist(),
                                     list(range(k)))

    def _linear(self, k, m):
        Y = np.arange(m) % k
        a = np.linspace(0, 2 * np.pi, m)
        X = np.stack([np.cos(a) + 10 * Y, np.sin(a)], 1)
        centroids = pvml.kmeans_train(X, k)
        Z, _ = pvml.kmeans_inference(X, centroids)
        self.assertListEqual(normalize_labels(Y).tolist(),
                             normalize_labels(Z).tolist())

    def test_linear_clusters(self):
        for k in range(1, 5):
            with self.subTest(k):
                self._linear(k, k)

    def test_respawn(self):
        m = 100
        k = 5
        X = np.random.randn(m, 2)
        centroids = np.zeros((k, 2))
        centroids = pvml.kmeans_train(X, k, init_centroids=centroids)
        Z, _ = pvml.kmeans_inference(X, centroids)
        self.assertListEqual(np.unique(Z).tolist(), list(range(k)))

    def test_init(self):
        m = 100
        k = 5
        X = np.random.randn(m, 2)
        init_centroids = np.arange(k * 2).reshape(k, 2)
        centroids = pvml.kmeans_train(X, k, init_centroids=init_centroids, steps=0)
        self.assertListEqual(init_centroids.tolist(), centroids.tolist())

    def test_errors1(self):
        X = np.random.randn(10, 2)
        with self.assertRaises(ValueError):
            pvml.kmeans_train(X, 0)

    def test_errors2(self):
        X = np.random.randn(10, 2)
        with self.assertRaises(ValueError):
            pvml.kmeans_train(X, 3, init_centroids=np.zeros((2, 2)))

    def test_errors3(self):
        X = np.random.randn(10, 2)
        with self.assertRaises(ValueError):
            pvml.kmeans_train(X, 3, init_centroids=np.zeros((3, 1)))

    def test_errors4(self):
        X = np.random.randn(10, 2)
        with self.assertRaises(ValueError):
            pvml.kmeans_train(X, 11)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import unittest
import test_data


class TestGDA(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_heteroscedastic_gda(self):
        for k in range(1, 5):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(48, k)
                means, icovs, priors = pvml.hgda_train(X, Y)
                Yhat, scores = pvml.hgda_inference(X, means, icovs, priors)
                self.assertListEqual(Y.tolist(), Yhat.tolist())
                self.assertListEqual(priors.tolist(), [1.0 / k] * k)

    def test_heteroscedastic_priors(self):
        k = 3
        X, Y = test_data.non_separable_checkerboard_data_set(24, k)
        means, icovs, priors = pvml.hgda_train(X, Y, priors=np.array([1, 0, 0]))
        Yhat, scores = pvml.hgda_inference(X, means, icovs, priors)
        self.assertListEqual([0] * 24, Yhat.tolist())

    def test_omoscedastic_gda(self):
        for k in range(1, 5):
            with self.subTest(k):
                X, Y = test_data.separable_stripes_data_set(48, k)
                w, b = pvml.ogda_train(X, Y)
                Yhat, scores = pvml.ogda_inference(X, w, b)
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_omoscedastic_priors(self):
        k = 4
        X, Y = test_data.non_separable_checkerboard_data_set(24, k)
        w, b = pvml.ogda_train(X, Y, priors=np.array([1, 0, 0, 0]))
        Yhat, scores = pvml.ogda_inference(X, w, b)
        self.assertListEqual([0] * 24, Yhat.tolist())

    def test_mindist(self):
        for k in range(1, 5):
            with self.subTest(k):
                X, Y = test_data.separable_hypercubes_data_set(48, k)
                means = pvml.mindist_train(X, Y)
                Yhat, scores = pvml.mindist_inference(X, means)
                self.assertListEqual(Y.tolist(), Yhat.tolist())


if __name__ == '__main__':
    unittest.main()



import unittest
import pvml
import numpy as np


class TestPCA(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_pca1(self):
        a = 2 * np.pi * np.arange(50) / 50
        X = np.stack([2 + np.cos(a), -1 + np.sin(a)], 1)
        X1 = pvml.pca(X)
        self.assertAlmostEqual(X1.mean(0)[0], 0)
        self.assertAlmostEqual(X1.mean(0)[1], 0)
        self.assertAlmostEqual(X1.var(0)[0], X.var(0)[0])
        self.assertAlmostEqual(X1.var(0)[1], X.var(0)[1])

    def test_pca2(self):
        a = np.arange(10)
        X = np.tile(a, (7, 1))
        T = np.ones((10, 10))
        X1, T1 = pvml.pca(X, T)
        self.assertEqual(X1.shape[1], 1)
        self.assertAlmostEqual(X1.mean(0), 0)
        self.assertEqual(T1[:, 0].tolist(), [1] * 10)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import io
import unittest
import test_data


class MLPTanh(pvml.MLP):
    """MLP with hyperbolic tangent activation function.

    It is used here because tanh is more reliable than ReLU for
    numerical testing.
    """
    def forward_hidden_activation(self, X):
        """Activation function of hidden layers."""
        return np.tanh(X)

    def backward_hidden_activation(self, Y, d):
        """Derivative of the activation function of hidden layers."""
        return d * (1 - Y ** 2)


class TestMLP(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_parameter_gradients(self):
        """Test the gradient of the loss wrt the parameters."""
        m = 4
        n = 3
        k = 5
        net = MLPTanh([n, 2, 7, k])
        X = np.random.randn(m, n)
        Y = np.arange(m) % k
        A = net.forward(X)
        D = net.backward(Y, A)
        loss = net.loss(Y, A[-1])
        grad_Ws, grad_bs = net.parameters_gradient(A, D)
        eps = 1e-7
        for ps, grad_ps, name in ((net.weights, grad_Ws, "W"), (net.biases, grad_bs, "b")):
            for p, grad_p, l in zip(ps, grad_ps, range(len(ps))):
                for index in np.ndindex(*p.shape):
                    backup = p[index]
                    p[index] += eps
                    with self.subTest(parameter=f"{name}[{l}]", index=index):
                        A1 = net.forward(X)
                        loss1 = net.loss(Y, A1[-1])
                        ratio = (loss1 - loss) / eps
                        self.assertAlmostEqual(grad_p[index], ratio, 5)
                    p[index] = backup

    def test_backward(self):
        m = 3
        n = 2
        k = 3
        net = MLPTanh([n, 2, 3, 3, 2, k])
        X = np.random.randn(m, n)
        Y = np.arange(m) % k
        A = net.forward(X)
        loss = net.loss(Y, A[-1])
        D = net.backward(Y, A)
        grad = net.backward_to_input(D[0])
        eps = 1e-7
        for index in np.ndindex(*X.shape):
            backup = X[index]
            X[index] += eps
            with self.subTest(index=index):
                A1 = net.forward(X)
                loss1 = net.loss(Y, A1[-1])
                ratio = (loss1 - loss) / eps
                self.assertAlmostEqual(grad[index], ratio, 5)
            X[index] = backup

    def test_train(self):
        X, Y = test_data.separable_hypercubes_data_set(50, 2)
        net = pvml.MLP([X.shape[1], 2])
        net.train(X, Y, lr=1e-1, steps=1000)
        Yhat, P = net.inference(X)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_batch(self):
        X, Y = test_data.separable_circle_data_set(50, 2)
        net = pvml.MLP([X.shape[1], 2, 2])
        net.train(X, Y, lr=1e-1, steps=1000, batch=10)
        Yhat, P = net.inference(X)
        self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_saveload(self):
        neurons = [2, 3, 4, 5]
        net1 = pvml.MLP(neurons)
        f = io.BytesIO()
        net1.save(f)
        f.seek(0)
        net2 = pvml.MLP.load(f)
        for w1, w2 in zip(net1.weights, net2.weights):
            self.assertListEqual(list(w1.ravel()), list(w2.ravel()))
        for b1, b2 in zip(net1.biases, net2.biases):
            self.assertListEqual(list(b1), list(b2))


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import unittest
import test_data


class TestMulticlassKSVM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_train_ovo(self):
        for k in range(2, 6):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(50, k)
                a, b = pvml.one_vs_one_ksvm_train(X, Y, "rbf", 0.1, 0, lr=1e-1, steps=1000)
                Yhat = pvml.one_vs_one_ksvm_inference(X, X, a, b, "rbf", 0.1)[0]
                self.assertListEqual(Y.tolist(), Yhat.tolist())

    def test_train_ovr(self):
        for k in range(2, 6):
            with self.subTest(k):
                X, Y = test_data.separable_circle_data_set(50, k)
                a, b = pvml.one_vs_rest_ksvm_train(X, Y, "rbf", 0.1, 0, lr=1e-1, steps=1000)
                Yhat = pvml.one_vs_rest_ksvm_inference(X, X, a, b, "rbf", 0.1)[0]
                self.assertListEqual(Y.tolist(), Yhat.tolist())


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .utils import log_nowarn
from .checks import _check_size, _check_labels


def logreg_inference(X, w, b):
    """Predict class probabilities.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    w : ndarray, shape (n,)
         weight vector.
    b : float
         scalar bias.

    Returns
    -------
    ndarray, shape (m,)
        probability estimates (one per feature vector).
    """
    _check_size("mn, n, *", X, w, b)
    logits = X @ w + b
    return sigmoid(logits)


def binary_cross_entropy(Y, P):
    """Average cross entropy.

    Parameters
    ----------
    Y : ndarray, shape (m,)
        binary target labels (0 or 1).
    P : ndarray, shape (m,)
        probability estimates.

    Returns
    -------
    float
        average cross entropy.
    """
    _check_size("m, m", Y, P)
    Y = _check_labels(Y, 2)
    log1 = log_nowarn(P)
    log0 = log_nowarn(1 - P)
    e = -log1[Y == 1].sum() - log0[Y == 0].sum()
    return e / Y.size


def logreg_train(X, Y, lambda_, lr=1e-3, steps=1000, init_w=None,
                 init_b=0):
    """Train a binary classifier based on L2-regularized logistic regression.

    Parameters
    ----------

    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate.
    steps : int
        number of training steps.
    init_w : ndarray, shape (n,)
        initial weights (None for zero initialization)
    init_b : float
        initial bias

    Returns
    -------

    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    _check_size("mn, m, n?, *", X, Y, init_w, init_b)
    Y = _check_labels(Y, 2)
    m, n = X.shape
    w = (init_w if init_w is not None else np.zeros(n))
    b = init_b
    for step in range(steps):
        P = logreg_inference(X, w, b)
        grad_w = ((P - Y) @ X) / m + 2 * lambda_ * w
        grad_b = (P - Y).mean()
        w -= lr * grad_w
        b -= lr * grad_b
    return w, b


def logreg_l1_train(X, Y, lambda_, lr=1e-3, steps=1000, init_w=None, init_b=0):
    """Train a binary classifier based on L1-regularized logistic regression.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate.
    steps : int
        number of training steps.
    loss : ndarray, shape (steps,)
        loss value after each training step.

    Returns
    -------
    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    _check_size("mn, m, n?, *", X, Y, init_w, init_b)
    Y = _check_labels(Y, 2)
    m, n = X.shape
    w = (init_w if init_w is not None else np.zeros(n))
    b = init_b
    for step in range(steps):
        P = logreg_inference(X, w, b)
        grad_w = ((P - Y) @ X) / m + lambda_ * np.sign(w)
        grad_b = (P - Y).mean()
        w -= lr * grad_w
        b -= lr * grad_b
    return w, b


def sigmoid(z):
    """Elementwise .

    Parameters
    ----------
    z : ndarray
         input

    Returns
    -------
    ndarray, (same shape of z)
        the sigmoid of z
    """
    return 1 / (1 + np.exp(-z))



import numpy as np
import unittest


def separable_circle_data_set(n, k):
    ii = np.arange(n)
    a = (np.pi / 5) + 2 * np.pi * ii / n
    Y = (k * ii) // n
    X = np.stack([np.cos(a) + 0.5, np.sin(a) - 0.5], 1)
    return X, Y


def separable_hypercubes_data_set(n, k):
    X = np.random.random((n, k))
    Y = np.arange(n) % k
    X[np.arange(n), Y] += Y
    return X, Y


def separable_stripes_data_set(n, k):
    rs = np.sqrt(n).astype(int)
    r = np.arange(n) // rs
    c = np.arange(n) % rs
    v1 = np.array([1, 0.1])
    v2 = np.array([-0.1, 1])
    X = np.outer(r, v1) + np.outer(c, v2)
    Y = (k * r) // (r.max() + 1)
    return X, Y


def non_separable_checkerboard_data_set(n, k):
    rs = np.sqrt(n).astype(int)
    r = np.arange(n) // rs
    c = np.arange(n) % rs
    v1 = np.array([0.7, -0.2])
    v2 = np.array([0.2, -0.7])
    X = np.outer(r, v1) + np.outer(c, v2)
    Y = (r + c) % k
    return X, Y


def categorical_data_set(n, k):
    Y = np.arange(n) % k
    X = Y[:, None] % np.array([2, 3, k])
    return X, Y


def bow_data_set(n, k):
    Y = np.arange(n) % k
    Z = np.arange(k * 3) // 3
    P = 0.1 + 0.8 * (Z[None, :, None] == Y[:, None, None])
    X = (np.random.random((n, k * 3, 5)) < P).sum(2)
    return X, Y


class TestTestData(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_separable_circle(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = separable_circle_data_set(4 * k, k)
                c = np.bincount(Y, minlength=k)
                self.assertEqual(c.min(), c.max())

    def test_separable_hypercubes(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = separable_hypercubes_data_set(4 * k, k)
                c = np.bincount(Y, minlength=k)
                self.assertEqual(c.min(), c.max())

    def test_separable_stripes(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = separable_stripes_data_set(4 * k, k)
                self.assertEqual(Y.min(), 0)
                self.assertEqual(Y.max(), k - 1)

    def test_non_separable_checkerboard(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = non_separable_checkerboard_data_set(4 * k, k)
                self.assertEqual(Y.min(), 0)
                self.assertEqual(Y.max(), k - 1)

    def test_categorical(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = categorical_data_set(4 * k, k)
                self.assertEqual(Y.min(), 0)
                self.assertEqual(Y.max(), k - 1)
                self.assertEqual(np.abs(np.modf(X)[0]).max(), 0)

    def test_bow(self):
        for k in range(1, 6):
            with self.subTest(k):
                X, Y = bow_data_set(4 * k, k)
                self.assertEqual(Y.min(), 0)
                self.assertEqual(Y.max(), k - 1)
                self.assertEqual(np.abs(np.modf(X)[0]).max(), 0)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pvml
import io
import unittest


class TestRNNBaseCell(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_gradientW(self):
        """Test the gradient of the loss wrt the parameters.

        The loss here is just the sum of all the cell states.
        """
        n, h, m, t = 3, 2, 4, 5
        init = np.zeros((m, h))
        cell = pvml.RNNBasicCell(n, h)
        X = np.random.randn(m, t, n)
        H = cell.forward(X, np.zeros((m, h)))
        L = H.sum()
        cell.backward(np.ones_like(H))
        gradients = cell.parameters_grad()
        eps = 1e-7
        for p in range(len(gradients)):
            for index in np.ndindex(*gradients[p].shape):
                backup = cell.parameters()[p][index]
                with self.subTest(parameter=p, index=index):
                    cell.parameters()[p][index] += eps
                    H1 = cell.forward(X, init)
                    L1 = H1.sum()
                    D = (L1 - L) / eps
                    self.assertAlmostEqual(gradients[p][index], D, 5)
                cell.parameters()[p][index] = backup


class TestRNN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_gradient_single_label(self):
        neurons = [3, 2, 4, 2, 3]
        m, t = 3, 5
        rnn = pvml.RNN(neurons)
        X = np.random.randn(m, t, neurons[0])
        Y = np.arange(m) % neurons[-1]
        P = rnn.forward(X)
        L = rnn.loss(Y, P)
        DX = rnn.backward(Y)
        eps = 1e-7
        for index in np.ndindex(*X.shape):
            backup = X[index]
            with self.subTest(index=index):
                X[index] += eps
                P1 = rnn.forward(X)
                L1 = rnn.loss(Y, P1)
                D = (L1 - L) / eps
                self.assertAlmostEqual(DX[index], D, 5)
            X[index] = backup

    def test_gradient_sequence(self):
        neurons = [3, 2, 4, 2, 3]
        m, t = 3, 5
        rnn = pvml.RNN(neurons)
        X = np.random.randn(m, t, neurons[0])
        Y = np.arange(m * t).reshape(m, t) % neurons[-1]
        P = rnn.forward(X)
        L = rnn.loss(Y, P)
        DX = rnn.backward(Y)
        eps = 1e-7
        for index in np.ndindex(*X.shape):
            backup = X[index]
            with self.subTest(index=index):
                X[index] += eps
                P1 = rnn.forward(X)
                L1 = rnn.loss(Y, P1)
                D = (L1 - L) / eps
                self.assertAlmostEqual(DX[index], D, 5)
            X[index] = backup

    def test_train(self):
        k = 4
        neurons = [k, 2 * k, k]
        # Train to echo output a set sequence
        Y = np.arange(k).reshape(1, k)
        X = np.eye(k).reshape(1, k, k)
        rnn = pvml.RNN(neurons)
        rnn.train(X, Y, lr=0.01, steps=100)
        P = rnn.forward(X)
        Z = P[0, :, :].argmax(-1)
        self.assertListEqual(list(Z), list(Y[0, :]))

    def test_saveload(self):
        neurons = [2, 3, 4, 5]
        rnn = pvml.RNN(neurons)
        f = io.BytesIO()
        rnn.save(f)
        f.seek(0)
        rnn2 = pvml.RNN.load(f)
        self.assertListEqual(list(rnn.b), list(rnn2.b))


class TestGRUCell(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(7477)

    def test_forward(self):
        n, h, m, t = 5, 3, 2, 4
        init = np.zeros((m, h))
        cell = pvml.GRUCell(n, h)
        X = np.random.randn(m, t, n)
        H = cell.forward(X, init)
        self.assertEqual(H.shape, (m, t, h))

    def test_gradientX(self):
        n, h, m, t = 2, 2, 2, 3
        init = np.zeros((m, h))
        cell = pvml.GRUCell(n, h)
        X = np.random.randn(m, t, n)
        H = cell.forward(X, init)
        DL = np.ones_like(H)
        DX = cell.backward(DL)
        L0 = H.sum()
        eps = 1e-7
        for index in np.ndindex(*X.shape):
            backup = X[index]
            X[index] += eps
            with self.subTest(index=index):
                H1 = cell.forward(X, init)
                L1 = H1.sum()
                D = (L1 - L0) / eps
                self.assertAlmostEqual(DX[index], D, 5)
            X[index] = backup

    def test_gradientW(self):
        """Test the gradient of the loss wrt the parameters.

        The loss here is just the sum of all the cell states.
        """
        n, h, m, t = 3, 2, 4, 5
        init = np.zeros((m, h))
        cell = pvml.GRUCell(n, h)
        X = np.random.randn(m, t, n)
        H = cell.forward(X, np.zeros((m, h)))
        L = H.sum()
        cell.backward(np.ones_like(H))
        gradients = cell.parameters_grad()
        eps = 1e-7
        for p in range(len(gradients)):
            for index in np.ndindex(*gradients[p].shape):
                backup = cell.parameters()[p][index]
                with self.subTest(parameter=p, index=index):
                    cell.parameters()[p][index] += eps
                    H1 = cell.forward(X, init)
                    L1 = H1.sum()
                    D = (L1 - L) / eps
                    self.assertAlmostEqual(gradients[p][index], D, 5)
                cell.parameters()[p][index] = backup


class TestRNNAbstractCell(unittest.TestCase):
    def test_forward(self):
        n, h, m, t = 3, 2, 4, 5
        init = np.zeros((m, h))
        cell = pvml.RNNAbstractCell()
        X = np.zeros((m, t, n))
        with self.assertRaises(NotImplementedError):
            cell.forward(X, init)

    def test_backward(self):
        n, m, t = 3, 4, 5
        D = np.zeros((m, t, n))
        cell = pvml.RNNAbstractCell()
        with self.assertRaises(NotImplementedError):
            cell.backward(D)

    def test_parameters(self):
        cell = pvml.RNNAbstractCell()
        with self.assertRaises(NotImplementedError):
            cell.parameters()

    def test_parameters_grad(self):
        cell = pvml.RNNAbstractCell()
        with self.assertRaises(NotImplementedError):
            cell.parameters_grad()


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .svm import svm_train
from .checks import _check_size, _check_labels


def one_vs_one_svm_inference(X, W, b):
    """Multiclass linear SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    W : ndarray, shape (n, s)
         weights (one vector per pair of classes).
    b : ndarray, shape (s,)
         vector of biases.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    _check_size("mn, ns, s", X, W, b)
    # 1) recover the number of classes from s = 1 + 2 + ... + k
    m = X.shape[0]
    s = b.size
    k = int(1 + np.sqrt(1 + 8 * s)) // 2
    votes = np.zeros((m, k))
    logits = X @ W + b.T
    bin_labels = (logits > 0)
    # For each pair of classes...
    j = 0
    for pos in range(k):
        for neg in range(pos + 1, k):
            votes[:, pos] += bin_labels[:, j]
            votes[:, neg] += (1 - bin_labels[:, j])
            j += 1
    labels = np.argmax(votes, 1)
    return labels, votes


def one_vs_one_svm_train(X, Y, lambda_, lr=1e-3, steps=1000,
                         init_w=None, init_b=None):
    """Train a multi-class linear SVM using the one vs. one strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels in the range 0, ..., (k - 1).
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_w : ndarray, shape (n, k * (k - 1) // 2)
        initial weights (None for zero initialization)
    init_b : ndarray, shape (k * (k - 1) // 2,)
        initial biases (None for zero initialization)

    Returns
    -------
    W : ndarray, shape (n, k * (k - 1) // 2)
        learned weights (one vector per each pair of classes).
    b : ndarray, shape (k * (k - 1) // 2,)
        vector of biases.
    """
    _check_size("mn, m", X, Y)
    Y = _check_labels(Y)
    k = Y.max() + 1
    m, n = X.shape
    W = np.zeros((n, k * (k - 1) // 2))
    b = np.zeros(k * (k - 1) // 2)
    j = 0
    # For each pair of classes...
    for pos in range(k):
        for neg in range(pos + 1, k):
            # Build a training subset
            subset = (np.logical_or(Y == pos, Y == neg)).nonzero()[0]
            Xbin = X[subset, :]
            Ybin = (Y[subset] == pos)
            w1 = (None if init_w is None else init_w[:, j])
            b1 = (0 if init_b is None else init_b[j])
            # Train the classifier
            Wbin, bbin = svm_train(Xbin, Ybin, lambda_, lr=lr, steps=steps,
                                   init_w=w1, init_b=b1)
            W[:, j] = Wbin
            b[j] = bbin
            j += 1
    return W, b


def one_vs_rest_svm_inference(X, W, b):
    """Multiclass linear SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    W : ndarray, shape (n, k)
         weights (one vector per class).
    b : ndarray, shape (k,)
         vector of biases.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    _check_size("mn, nk, k", X, W, b)
    logits = X @ W + b.T
    labels = np.argmax(logits, 1)
    return labels, logits


def one_vs_rest_svm_train(X, Y, lambda_, lr=1e-3, steps=1000,
                          init_w=None, init_b=None):
    """Train a multi-class linear SVM using the one vs. rest strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels in the range 0, ..., (k - 1).
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_w : ndarray, shape (n, k)
        initial weights (None for zero initialization)
    init_b : ndarray, shape (k,)
        initial biases (None for zero initialization)

    Returns
    -------
    W : ndarray, shape (n, k)
        learned weights (one vector per class).
    b : ndarray, shape (k,)
        vector of biases.
    """
    _check_size("mn, m", X, Y)
    Y = _check_labels(Y)
    k = Y.max() + 1
    m, n = X.shape
    W = np.zeros((n, k))
    b = np.zeros(k)
    for c in range(k):
        Ybin = (Y == c)
        w1 = (None if init_w is None else init_w[:, c])
        b1 = (0 if init_b is None else init_b[c])
        Wbin, bbin = svm_train(X, Ybin, lambda_, lr=lr, steps=steps,
                               init_w=w1, init_b=b1)
        W[:, c] = Wbin
        b[c] = bbin
    return W, b



import numpy as np
from .gda import mindist_inference
from .checks import _check_size


def kmeans_train(X, k, steps=1000, init_centroids=None):
    """K-means clustering algorithm.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    k : int
        number of clusters.
    steps : int
        maximum number of iterations.
    init_centroids : ndarray, shape (k, n)
        initial centroid (None for random initialization).

    Returns
    -------
    centroids : ndarray, shape (k, n)
         class centroids.
    """
    _check_size("mn, kn?", X, init_centroids)
    _check_centroids(X, k, init_centroids)
    m, n = X.shape
    # Initialization
    if init_centroids is None:
        centroids = np.empty((k, n))
        Y = np.arange(m) % k
        np.random.shuffle(Y)
    else:
        centroids = init_centroids
        Y, _ = kmeans_inference(X, centroids)
    # Main loop
    for step in range(steps):
        # Update the centroids
        counts = np.bincount(Y, minlength=k)
        while (counts == 0).any():
            _respawn_empty(Y, counts)
        for i in range(k):
            centroids[i, :] = X[Y == i, :].mean(0)
        # Update the labels
        Yold = Y
        Y, _ = kmeans_inference(X, centroids)
        # Stop if nothing changed
        if (Y == Yold).all():
            break
    return centroids


def kmeans_inference(X, centroids):
    """K-means inference.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    centroids : ndarray, shape (k, n)
         class mean vectors.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m, k)
        scores assigned to each class.
    """
    # After training K-means is just a minimum distance classifier
    return mindist_inference(X, centroids)


def _respawn_empty(Y, counts):
    # Divide the samples in the largest cluster to replace the first empty one.
    big = np.argmax(counts)
    empty = (counts == 0).nonzero()[0][0]
    Ynew = np.random.choice(np.array([empty, big]), (counts[big],))
    changes = (Ynew == empty).sum()
    Y[Y == big] = Ynew
    counts[empty] = changes
    counts[big] -= changes


def _check_centroids(X, k, centroids):
    if k < 1:
        msg = "The number of clusters ({}) must be a strictly positive integer"
        raise ValueError(msg.format(k))
    if centroids is not None and k != centroids.shape[0]:
        msg = "The number of clusters ({}) "
        msg += "does not match the number of centroids ({})"
        raise ValueError(msg.format(k, centroids.shape[0]))
    if k > X.shape[0]:
        msg = "The number of samples ({}) "
        msg += "is smaller than the number of clusters ({})"
        raise ValueError(msg.format(X.shape[0], k))
